# Owner(s): ["module: nn"]
import math
import random
import string
import unittest
from functools import reduce
from operator import mul

import torch
import torch.nn.functional as F
import torch.nn.init as init
from torch.testing._internal.common_device_type import instantiate_device_type_tests
from torch.testing._internal.common_utils import (
    parametrize as parametrize_test,
    run_tests,
    skipIfNoLapack,
    slowTest,
    TEST_SCIPY,
    TestCase,
)


if TEST_SCIPY:
    from scipy import stats


# (init_fn, shape, extra kwargs) covering every public init function.
# Shapes satisfy each function's dimensionality requirements.
# The 0-element variant is derived by replacing dim-0 with 0.
ALL_INIT_FNS = [
    (init.uniform_, (3, 5), {}),
    (init.normal_, (3, 5), {}),
    (init.trunc_normal_, (3, 5), {}),
    (init.constant_, (3, 5), {"val": 0.3}),
    (init.ones_, (3, 5), {}),
    (init.zeros_, (3, 5), {}),
    (init.eye_, (3, 5), {}),
    (init.dirac_, (3, 16, 5), {}),
    (init.xavier_uniform_, (3, 5), {}),
    (init.xavier_normal_, (3, 5), {}),
    (init.kaiming_uniform_, (3, 5), {}),
    (init.kaiming_normal_, (3, 5), {}),
    (init.orthogonal_, (3, 5), {}),
    (init.sparse_, (3, 5), {"sparsity": 0.1}),
]


class TestNNInit(TestCase):
    def setUp(self):
        super().setUp()
        random.seed(123)

    @torch._dynamo.disable
    def _is_normal(self, tensor, mean, std):
        samples = tensor.view(-1).tolist()
        p_value = stats.kstest(samples, "norm", args=(mean, std))[1]
        return p_value > 0.0001

    @torch._dynamo.disable
    def _is_trunc_normal(self, tensor, mean, std, a, b):
        # scipy's trunc norm is suited for data drawn from N(0, 1),
        # so we need to transform our data to test it using scipy.
        z_samples = (tensor.view(-1) - mean) / std
        z_samples = z_samples.tolist()
        a0 = (a - mean) / std
        b0 = (b - mean) / std
        p_value = stats.kstest(z_samples, "truncnorm", args=(a0, b0))[1]
        return p_value > 0.0001

    @torch._dynamo.disable
    def _is_uniform(self, tensor, a, b):
        samples = tensor.view(-1).tolist()
        p_value = stats.kstest(samples, "uniform", args=(a, (b - a)))[1]
        return p_value > 0.0001

    def _create_random_nd_tensor(self, dims, size_min, size_max):
        size = [random.randint(size_min, size_max) for _ in range(dims)]
        tensor = torch.zeros(size)
        return tensor

    def _random_float(self, a, b):
        return (b - a) * random.random() + a

    def test_calculate_gain_linear(self):
        for fn in [
            "linear",
            "conv1d",
            "conv2d",
            "conv3d",
            "conv_transpose2d",
            "conv_transpose2d",
            "conv_transpose3d",
        ]:
            gain = init.calculate_gain(fn)
            self.assertEqual(gain, 1)

    def test_calculate_gain_nonlinear(self):
        for fn in ["sigmoid", "tanh", "relu", "leaky_relu"]:
            gain = init.calculate_gain(fn)
            if fn == "sigmoid":
                self.assertEqual(gain, 1)
            elif fn == "tanh":  # 5 / 3
                self.assertEqual(gain, 1.6666666666666667)
            elif fn == "relu":  # sqrt(2)
                self.assertEqual(gain, 1.4142135623730951)
            elif fn == "leaky_relu":  # sqrt(2 / 1 + slope^2))
                self.assertEqual(gain, 1.4141428569978354)
            elif fn == "selu":
                self.assertEqual(gain, 0.75)

    def test_calculate_gain_leaky_relu(self):
        for param in [None, 0, 0.01, 10]:
            gain = init.calculate_gain("leaky_relu", param)
            if param is None:  # Default slope is 0.01
                self.assertEqual(gain, 1.4141428569978354)
            elif param == 0:  # No slope = same gain as normal ReLU
                self.assertEqual(gain, 1.4142135623730951)
            elif param == 0.01:
                self.assertEqual(gain, 1.4141428569978354)
            elif param == 10:
                self.assertEqual(gain, 0.14071950894605836)

    def test_calculate_gain_leaky_relu_only_accepts_numbers(self):
        for param in [True, [1], {"a": "b"}]:
            with self.assertRaises(ValueError):
                init.calculate_gain("leaky_relu", param)

    def test_calculate_gain_only_accepts_valid_nonlinearities(self):
        for n in [2, 5, 25]:
            # Generate random strings of lengths that definitely aren't supported
            random_string = "".join(
                [random.choice(string.ascii_lowercase) for i in range(n)]
            )
            with self.assertRaises(ValueError):
                init.calculate_gain(random_string)

    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    def test_uniform(self):
        for dims in [1, 2, 4]:
            input_tensor = self._create_random_nd_tensor(dims, size_min=30, size_max=50)
            a = self._random_float(-3, 3)
            b = a + self._random_float(1, 5)
            init.uniform_(input_tensor, a=a, b=b)
            if not self._is_uniform(input_tensor, a, b):
                raise AssertionError("Expected uniform distribution")

    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    def test_normal(self):
        for dims in [1, 2, 4]:
            input_tensor = self._create_random_nd_tensor(dims, size_min=30, size_max=50)
            mean = self._random_float(-3, 3)
            std = self._random_float(1, 5)
            init.normal_(input_tensor, mean=mean, std=std)

            if not self._is_normal(input_tensor, mean, std):
                raise AssertionError("Expected normal distribution")

    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    def test_trunc_normal_generator(self):
        gen = torch.Generator()
        gen.manual_seed(42)
        input_tensor = torch.empty(5)
        init.trunc_normal_(input_tensor, generator=gen)

        ref = torch.empty(5)
        torch.manual_seed(42)
        init.trunc_normal_(ref)

        self.assertEqual(input_tensor, ref)
        if not self._is_trunc_normal(input_tensor, mean=0, std=1, a=0, b=1):
            raise AssertionError("Expected truncated normal distribution")

    def test_init_methods_meta(self):
        for fn, shape, kwargs in ALL_INIT_FNS:
            with self.subTest(init_fn=fn.__name__):
                t = torch.empty(shape, device="meta")
                result = fn(t, **kwargs)
                self.assertIs(result, t)

    def test_init_methods_meta_zero_element(self):
        for fn, shape, kwargs in ALL_INIT_FNS:
            with self.subTest(init_fn=fn.__name__):
                t = torch.empty((0,) + shape[1:], device="meta")
                result = fn(t, **kwargs)
                self.assertIs(result, t)

    def test_constant(self):
        for dims in [1, 2, 4]:
            input_tensor = self._create_random_nd_tensor(dims, size_min=1, size_max=5)
            val = self._random_float(1, 10)
            init.constant_(input_tensor, val)

            self.assertEqual(input_tensor, input_tensor.clone().fill_(val))

    def test_ones_and_zeros(self):
        for init_fn_, val in zip([init.ones_, init.zeros_], [1, 0]):
            for dims in [1, 2, 4]:
                input_tensor = self._create_random_nd_tensor(
                    dims, size_min=1, size_max=5
                )
                init_fn_(input_tensor)

                self.assertEqual(input_tensor, input_tensor.clone().fill_(val))

    def test_eye(self):
        input_tensor = self._create_random_nd_tensor(2, size_min=1, size_max=5)
        init.eye_(input_tensor)

        # Check every single element
        for i in range(input_tensor.size(0)):
            for j in range(input_tensor.size(1)):
                if i == j:
                    if input_tensor[i][j] != 1:
                        raise AssertionError(
                            f"Expected 1 at [{i}][{j}], got {input_tensor[i][j]}"
                        )
                else:
                    if input_tensor[i][j] != 0:
                        raise AssertionError(
                            f"Expected 0 at [{i}][{j}], got {input_tensor[i][j]}"
                        )

    def test_eye_only_works_on_2d_inputs(self):
        for dims in [1, 3]:
            with self.assertRaises(ValueError):
                tensor = self._create_random_nd_tensor(dims, size_min=1, size_max=3)
                init.eye_(tensor)

    def test_dirac_properties(self):
        for dims in [3, 4, 5]:
            for groups in [1, 2, 3]:
                # prepare random tensor with random sizes, but fits groups
                a, c, d, e = (random.randint(1, 5) for _ in range(4))
                b = random.randint(
                    1, 5 * groups
                )  # same range as a*groups but all range allowed
                # make sure first dim divides by groups
                input_tensor = torch.randn((a * groups, b, c, d, e)[:dims])

                init.dirac_(input_tensor, groups)

                c_out, c_in = input_tensor.size(0) // groups, input_tensor.size(1)
                min_d = min(c_out, c_in)
                # Check number of nonzeros is equivalent to smallest dim (for each group)
                if torch.nonzero(input_tensor).size(0) != min_d * groups:
                    raise AssertionError(
                        f"Expected {min_d * groups} nonzeros, got {torch.nonzero(input_tensor).size(0)}"
                    )
                # Check sum of values (can have precision issues, hence assertEqual) is also equivalent
                self.assertEqual(input_tensor.sum(), min_d * groups)

    def test_dirac_identity(self):
        for groups in [1, 3]:
            batch, in_c, out_c, size, kernel_size = (
                8,
                3,
                9,
                5,
                3,
            )  # in_c, out_c must divide by groups
            eff_out_c = out_c // groups

            # Test 1D
            input_var = torch.randn(batch, in_c, size)
            filter_var = torch.zeros(eff_out_c, in_c, kernel_size)
            filter_var = torch.cat([filter_var] * groups)
            init.dirac_(filter_var, groups)
            output_var = F.conv1d(input_var, filter_var)
            input_tensor, output_tensor = (
                input_var.data,
                output_var.data,
            )  # Variables do not support nonzero
            for g in range(groups):
                # Assert in_c outputs are preserved (per each group)
                self.assertEqual(
                    input_tensor[:, :, 1:-1],
                    output_tensor[:, eff_out_c * g : eff_out_c * g + in_c, :],
                )
                # Assert extra outputs are 0
                if (
                    torch.nonzero(
                        output_tensor[:, eff_out_c * g + in_c : eff_out_c * (g + 1), :]
                    ).numel()
                    != 0
                ):
                    raise AssertionError("Expected extra outputs to be 0")

            # Test 2D
            input_var = torch.randn(batch, in_c, size, size)
            filter_var = torch.zeros(eff_out_c, in_c, kernel_size, kernel_size)
            filter_var = torch.cat([filter_var] * groups)
            init.dirac_(filter_var, groups)
            output_var = F.conv2d(input_var, filter_var)
            input_tensor, output_tensor = (
                input_var.data,
                output_var.data,
            )  # Variables do not support nonzero
            for g in range(groups):
                # Assert in_c outputs are preserved (per each group)
                self.assertEqual(
                    input_tensor[:, :, 1:-1, 1:-1],
                    output_tensor[:, eff_out_c * g : eff_out_c * g + in_c, :, :],
                )
                # Assert extra outputs are 0
                if (
                    torch.nonzero(
                        output_tensor[
                            :, eff_out_c * g + in_c : eff_out_c * (g + 1), :, :
                        ]
                    ).numel()
                    != 0
                ):
                    raise AssertionError("Expected extra outputs to be 0")

            # Test 3D
            input_var = torch.randn(batch, in_c, size, size, size)
            filter_var = torch.zeros(
                eff_out_c, in_c, kernel_size, kernel_size, kernel_size
            )
            filter_var = torch.cat([filter_var] * groups)
            init.dirac_(filter_var, groups)
            output_var = F.conv3d(input_var, filter_var)
            input_tensor, output_tensor = input_var.data, output_var.data
            for g in range(groups):
                # Assert in_c outputs are preserved (per each group)
                self.assertEqual(
                    input_tensor[:, :, 1:-1, 1:-1, 1:-1],
                    output_tensor[:, eff_out_c * g : eff_out_c * g + in_c, :, :, :],
                )
                # Assert extra outputs are 0
                if (
                    torch.nonzero(
                        output_tensor[
                            :, eff_out_c * g + in_c : eff_out_c * (g + 1), :, :, :
                        ]
                    ).numel()
                    != 0
                ):
                    raise AssertionError("Expected extra outputs to be 0")

    def test_dirac_only_works_on_3_4_5d_inputs(self):
        for dims in [1, 2, 6]:
            with self.assertRaises(ValueError):
                tensor = self._create_random_nd_tensor(dims, size_min=1, size_max=3)
                init.dirac_(tensor)

    def test_xavier_uniform_errors_on_inputs_smaller_than_2d(self):
        for dims in [0, 1]:
            tensor = self._create_random_nd_tensor(dims, size_min=1, size_max=1)
            with self.assertRaises(ValueError):
                init.xavier_uniform_(tensor)

    def test_xavier_normal_errors_on_inputs_smaller_than_2d(self):
        for dims in [0, 1]:
            tensor = self._create_random_nd_tensor(dims, size_min=1, size_max=1)
            with self.assertRaises(ValueError):
                init.xavier_normal_(tensor)

    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    @slowTest
    def test_xavier_uniform(self):
        for use_gain in [True, False]:
            for dims in [2, 4]:
                input_tensor = self._create_random_nd_tensor(
                    dims, size_min=20, size_max=25
                )
                gain = 1

                if use_gain:
                    gain = self._random_float(0.1, 2)
                    init.xavier_uniform_(input_tensor, gain=gain)
                else:
                    init.xavier_uniform_(input_tensor)

                fan_in = input_tensor.size(1)
                fan_out = input_tensor.size(0)
                if input_tensor.dim() > 2:
                    fan_in *= input_tensor[0, 0].numel()
                    fan_out *= input_tensor[0, 0].numel()

                expected_std = gain * math.sqrt(2.0 / (fan_in + fan_out))
                bounds = expected_std * math.sqrt(3)
                if not self._is_uniform(input_tensor, -bounds, bounds):
                    raise AssertionError("Expected uniform distribution")

    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    def test_xavier_normal(self):
        for use_gain in [True, False]:
            for dims in [2, 4]:
                input_tensor = self._create_random_nd_tensor(
                    dims, size_min=20, size_max=25
                )
                gain = 1

                if use_gain:
                    gain = self._random_float(0.1, 2)
                    init.xavier_normal_(input_tensor, gain=gain)
                else:
                    init.xavier_normal_(input_tensor)

                fan_in = input_tensor.size(1)
                fan_out = input_tensor.size(0)
                if input_tensor.dim() > 2:
                    fan_in *= input_tensor[0, 0].numel()
                    fan_out *= input_tensor[0, 0].numel()

                expected_std = gain * math.sqrt(2.0 / (fan_in + fan_out))
                if not self._is_normal(input_tensor, 0, expected_std):
                    raise AssertionError("Expected normal distribution")

    def test_kaiming_uniform_errors_on_inputs_smaller_than_2d(self):
        for dims in [0, 1]:
            with self.assertRaises(ValueError):
                tensor = self._create_random_nd_tensor(dims, size_min=1, size_max=1)
                init.kaiming_uniform_(tensor)

    def test_kaiming_normal_errors_on_inputs_smaller_than_2d(self):
        for dims in [0, 1]:
            with self.assertRaises(ValueError):
                tensor = self._create_random_nd_tensor(dims, size_min=1, size_max=1)
                init.kaiming_normal_(tensor)

    def test_kaiming_uniform_warning_on_0element_tensor(self):
        tensor = torch.empty(0, 1)
        with self.assertWarnsRegex(
            UserWarning, "Initializing zero-element tensors is a no-op"
        ):
            _ = init.kaiming_uniform_(tensor)

    def test_kaiming_normal_warning_on_0element_tensor(self):
        tensor = torch.empty(0, 1)
        with self.assertWarnsRegex(
            UserWarning, "Initializing zero-element tensors is a no-op"
        ):
            _ = init.kaiming_normal_(tensor)

    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    def test_kaiming_uniform(self):
        for use_a in [True, False]:
            for dims in [2, 4]:
                for mode in ["fan_in", "fan_out"]:
                    input_tensor = self._create_random_nd_tensor(
                        dims, size_min=20, size_max=25
                    )
                    if use_a:
                        a = self._random_float(0.1, 2)
                        init.kaiming_uniform_(input_tensor, a=a, mode=mode)
                    else:
                        a = 0
                        init.kaiming_uniform_(input_tensor, mode=mode)

                    fan_in = input_tensor.size(1)
                    fan_out = input_tensor.size(0)
                    if input_tensor.dim() > 2:
                        fan_in *= input_tensor[0, 0].numel()
                        fan_out *= input_tensor[0, 0].numel()

                    if mode == "fan_in":
                        n = fan_in
                    else:
                        n = fan_out

                    expected_std = math.sqrt(2.0 / ((1 + a**2) * n))
                    bounds = expected_std * math.sqrt(3.0)
                    if not self._is_uniform(input_tensor, -bounds, bounds):
                        raise AssertionError("Expected uniform distribution")

    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    def test_kaiming_normal(self):
        for use_a in [True, False]:
            for dims in [2, 4]:
                for mode in ["fan_in", "fan_out"]:
                    input_tensor = self._create_random_nd_tensor(
                        dims, size_min=20, size_max=25
                    )
                    if use_a:
                        a = self._random_float(0.1, 2)
                        init.kaiming_normal_(input_tensor, a=a, mode=mode)
                    else:
                        a = 0
                        init.kaiming_normal_(input_tensor, mode=mode)

                    fan_in = input_tensor.size(1)
                    fan_out = input_tensor.size(0)
                    if input_tensor.dim() > 2:
                        fan_in *= input_tensor[0, 0].numel()
                        fan_out *= input_tensor[0, 0].numel()

                    if mode == "fan_in":
                        n = fan_in
                    else:
                        n = fan_out

                    expected_std = math.sqrt(2.0 / ((1 + a**2) * n))
                    if not self._is_normal(input_tensor, 0, expected_std):
                        raise AssertionError("Expected normal distribution")

    def test_sparse_only_works_on_2d_inputs(self):
        for dims in [1, 3]:
            with self.assertRaises(ValueError):
                sparsity = self._random_float(0.1, 0.9)
                tensor = self._create_random_nd_tensor(dims, size_min=1, size_max=3)
                init.sparse_(tensor, sparsity)

    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    def test_sparse_default_std(self):
        for use_random_std in [True, False]:
            input_tensor = self._create_random_nd_tensor(2, size_min=30, size_max=35)
            rows = input_tensor.size(0)
            sparsity = self._random_float(0.1, 0.2)

            std = 0.01  # default std
            if use_random_std:
                std = self._random_float(0.01, 0.2)
                init.sparse_(input_tensor, sparsity=sparsity, std=std)
            else:
                init.sparse_(input_tensor, sparsity=sparsity)

            for col_idx in range(input_tensor.size(1)):
                column = input_tensor[:, col_idx]
                if column[column == 0].nelement() < math.ceil(sparsity * rows):
                    raise AssertionError("Expected more zero elements for sparsity")

            if not self._is_normal(input_tensor[input_tensor != 0], 0, std):
                raise AssertionError(
                    "Expected normal distribution for non-zero elements"
                )

    @skipIfNoLapack
    def test_orthogonal(self):
        for use_gain in [True, False]:
            for tensor_size in [[3, 4], [4, 3], [20, 2, 3, 4], [2, 3, 4, 5]]:
                input_tensor = torch.zeros(tensor_size)
                gain = 1.0

                if use_gain:
                    gain = self._random_float(0.1, 2)
                    init.orthogonal_(input_tensor, gain=gain)
                else:
                    init.orthogonal_(input_tensor)

                rows, cols = tensor_size[0], reduce(mul, tensor_size[1:])
                flattened_tensor = input_tensor.view(rows, cols)
                if rows > cols:
                    self.assertEqual(
                        torch.mm(flattened_tensor.t(), flattened_tensor),
                        torch.eye(cols) * gain**2,
                        atol=1e-6,
                        rtol=0,
                    )
                else:
                    self.assertEqual(
                        torch.mm(flattened_tensor, flattened_tensor.t()),
                        torch.eye(rows) * gain**2,
                        atol=1e-6,
                        rtol=0,
                    )

    def test_deprecation(self):
        x = torch.randn(3, 3)

        def fn():
            init.normal(x)

        with self.assertWarnsRegex(
            FutureWarning,
            "deprecated",
            msg="methods not suffixed with underscore should be deprecated",
        ):
            fn()


class TestNNInitDeviceType(TestCase):
    @torch._dynamo.disable
    def _is_trunc_normal(self, tensor, mean, std, a, b):
        z_samples = (tensor.view(-1) - mean) / std
        z_samples = z_samples.tolist()
        a0 = (a - mean) / std
        b0 = (b - mean) / std
        p_value = stats.kstest(z_samples, "truncnorm", args=(a0, b0))[1]
        return p_value > 0.0001

    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    @parametrize_test("dims", [1, 2, 4])
    @parametrize_test(
        "dtype",
        [torch.float32, torch.float64],
    )
    def test_trunc_normal_all_dtypes(self, device, dims, dtype):
        size = [random.randint(30, 50) for _ in range(dims)]
        input_tensor = torch.zeros(size, dtype=dtype, device=device)
        mean = random.uniform(-3, 3)
        std = random.uniform(0.01, 1)
        a = random.uniform(mean - 2 * std, mean)
        b = random.uniform(mean, mean + 2 * std)
        init.trunc_normal_(input_tensor, mean=mean, std=std, a=a, b=b)

        # Compare bounds at tensor precision: a and b get rounded when
        # cast to reduced-precision dtypes.
        lo = torch.tensor(a, dtype=dtype).item()
        hi = torch.tensor(b, dtype=dtype).item()
        self.assertTrue(
            input_tensor.min().item() >= lo,
            f"{dtype}: values below lower bound a={a}",
        )
        self.assertTrue(
            input_tensor.max().item() <= hi,
            f"{dtype}: values above upper bound b={b}",
        )
        self.assertTrue(
            self._is_trunc_normal(input_tensor.float().cpu(), mean, std, a, b),
            f"{dtype}: failed KS test against truncated normal",
        )

    # Reduced-precision KS test uses fixed wide params to avoid random
    # intervals too narrow for the type's representable value count.
    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    @parametrize_test(
        "dtype",
        [torch.float16, torch.bfloat16],
    )
    def test_trunc_normal_low_precision(self, device, dtype):
        n = 10000
        mean, std, a, b = 0.0, 1.0, -2.0, 2.0
        t = torch.empty(n, dtype=dtype, device=device)
        init.trunc_normal_(t, mean=mean, std=std, a=a, b=b)

        self.assertTrue(
            t.min().item() >= a,
            f"{dtype}: values below lower bound a={a}",
        )
        self.assertTrue(
            t.max().item() <= b,
            f"{dtype}: values above upper bound b={b}",
        )
        self.assertTrue(
            self._is_trunc_normal(t.float().cpu(), mean, std, a, b),
            f"{dtype}: failed KS test against truncated normal",
        )

    # Test that trunc_normal_ behaves well for narrow interval compared to std.
    @unittest.skipIf(not TEST_SCIPY, "Scipy not found.")
    @parametrize_test(
        "dtype",
        [torch.float32, torch.float64, torch.float16, torch.bfloat16],
    )
    def test_trunc_normal_narrow_interval(self, device, dtype):
        n = 10000
        mean, std, a, b = -3.0, 1.0, -0.5, 0.5
        t = torch.empty(n, dtype=dtype, device=device)
        init.trunc_normal_(t, mean=mean, std=std, a=a, b=b)

        self.assertTrue(
            t.min().item() >= a,
            f"{dtype}: values below lower bound a={a}",
        )
        self.assertTrue(
            t.max().item() <= b,
            f"{dtype}: values above upper bound b={b}",
        )
        self.assertTrue(
            self._is_trunc_normal(t.float().cpu(), mean, std, a, b),
            f"{dtype}: failed KS test against truncated normal",
        )

    # Sanity check for trunc normal to ensure that we sample a decent
    # subset of the value space.
    @parametrize_test(
        "dtype,min_unique",
        [
            (torch.float32, 5000),
            (torch.float64, 5000),
            (torch.float16, 3000),
            (torch.bfloat16, 1000),
        ],
    )
    def test_trunc_normal_unique_values(self, device, dtype, min_unique):
        n = 10000
        t = torch.empty(n, dtype=dtype, device=device)
        init.trunc_normal_(t, mean=0.0, std=1.0, a=-2.0, b=2.0)

        self.assertTrue(
            t.min().item() >= -2.0,
            f"{dtype}: values below lower bound",
        )
        self.assertTrue(
            t.max().item() <= 2.0,
            f"{dtype}: values above upper bound",
        )

        unique = t.unique().numel()
        self.assertGreater(
            unique,
            min_unique,
            f"{dtype}: only {unique} unique values, expected > {min_unique}",
        )

        self.assertFalse(
            t.isinf().any().item(),
            f"{dtype}: trunc_normal_ produced inf values",
        )
        self.assertFalse(
            t.isnan().any().item(),
            f"{dtype}: trunc_normal_ produced nan values",
        )

    # Sanity check that we don't round to the boundary by mistake.
    @parametrize_test(
        "dtype",
        [torch.float32, torch.float64, torch.float16, torch.bfloat16],
    )
    def test_trunc_normal_no_boundary_values_small_std(self, device, dtype):
        t = torch.empty(10000, dtype=dtype, device=device)
        init.trunc_normal_(t, mean=0.0, std=0.1, a=-2.0, b=2.0)

        at_lower = (t == -2.0).sum().item()
        at_upper = (t == 2.0).sum().item()
        self.assertEqual(
            at_lower,
            0,
            f"{dtype}: {at_lower} values clamped to lower bound a=-2.0",
        )
        self.assertEqual(
            at_upper,
            0,
            f"{dtype}: {at_upper} values clamped to upper bound b=2.0",
        )

    def _run_init_test(self, device, zero_element=False):
        for fn, shape, kwargs in ALL_INIT_FNS:
            if zero_element:
                shape = (0,) + shape[1:]
            with self.subTest(init_fn=fn.__name__):
                t = torch.empty(shape, device=device)
                result = fn(t, **kwargs)
                self.assertIs(result, t)

    def test_init_methods_no_error(self, device):
        # Smoke test: every public init function must run without error on all
        # devices (including meta where the ops are no-ops).
        self._run_init_test(device)

    def test_init_methods_zero_element(self, device):
        # Every init function should handle 0-element tensors as a no-op.
        self._run_init_test(device, zero_element=True)


instantiate_device_type_tests(TestNNInitDeviceType, globals())


if __name__ == "__main__":
    run_tests()



# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import math

import torch
from functorch.dim import cat, dimlists, dims
from torch import nn


class Linear(nn.Linear):
    def forward(self, input):
        ci, co = dims()
        b = dimlists()
        result = (input[b, ci] * self.weight[co, ci]).sum(ci) + self.bias[co]
        return result.order(b, co)


class BertSelfAttention(nn.Module):
    def __init__(
        self,
        hidden_size,
        num_attention_heads,
        attention_probs_dropout_prob,
        position_embedding_type=None,
        max_position_embeddings=None,
        linear=Linear,
    ):
        super().__init__()
        if hidden_size % num_attention_heads != 0:
            raise ValueError(
                f"The hidden size ({hidden_size}) is not a multiple of the number of attention "
                f"heads ({num_attention_heads})"
            )

        self.num_attention_heads = num_attention_heads
        self.attention_head_size = int(hidden_size / num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = linear(hidden_size, self.all_head_size)
        self.key = linear(hidden_size, self.all_head_size)
        self.value = linear(hidden_size, self.all_head_size)

        self.dropout_prob = attention_probs_dropout_prob
        self.position_embedding_type = position_embedding_type

        if self.position_embedding_type is not None:
            if max_position_embeddings is None:
                raise AssertionError(
                    "max_position_embeddings must not be None when "
                    f"position_embedding_type is {self.position_embedding_type}"
                )
            self.max_position_embeddings = max_position_embeddings
            self.distance_embedding = nn.Embedding(
                2 * max_position_embeddings - 1, self.attention_head_size
            )

    def forward(
        self,
        hidden_states,
        past_key_value=None,
    ):
        # first run the encoding linear layers for q, k, v normally
        # the meaning of a linear layer is well understood, so no need to use explicit dimensions
        q = self.query(hidden_states)
        k = self.key(hidden_states)
        v = self.value(hidden_states)

        # introduce values that represent each dimension. dimensions are 'first class'
        # because they are actual python values introduced here
        batch, query_sequence, key_sequence, heads, features = dims()
        heads.size = self.num_attention_heads

        # bind the positional dimensions in k, q, and v against
        # our values. the sizes of each dimension are determined by this binding
        # and when a dimension is used twice (e.g. batch), its size against both
        # uses is checked for consistency.
        # The group (heads, features) splits apart a single positional dimension
        # into two dimensions. Since heads.size*features.size == q.size(2)
        # and we specified heads.size, features.size is inferred here.
        q = q[batch, query_sequence, [heads, features]]
        k = k[batch, key_sequence, [heads, features]]
        v = v[batch, key_sequence, [heads, features]]

        # this option allows the model to attend to not just the elements of the current sequence
        # but the previous elements as well as additional tokens.
        if past_key_value is not None:
            extended_key_sequence = dims()
            key_past = past_key_value[0][batch, heads, key_sequence, features]
            value_past = past_key_value[1][batch, heads, key_sequence, features]
            # cat introduces a new dimension extended_key_sequence, because it is twice as long
            # as the original key_sequence
            k = cat([key_past, k], key_sequence, extended_key_sequence)
            v = cat([value_past, v], key_sequence, extended_key_sequence)
            # for the rest of the function, we will just use extended_key_sequence in lieu of
            # key_sequence
            key_sequence = extended_key_sequence

        # Take the dot product between "query" and "key" to get the raw attention scores.
        # The actual outer-product and summation are explicitly represented here,
        # and like einsum, will be pattern matched to an efficient matrix multiply op.
        attention_scores = (q * k).sum(features) / math.sqrt(features.size)

        # relative positional embeddings gave a unique embedding based on the distance between
        # key and value tokens in the sequence, e.g.
        #  0  1  2  3
        # -1  0  1  2
        # -2 -1  0  1
        # -3 -2 -1  0
        if self.position_embedding_type is not None:
            # the value of a dimension object when used as a tensor is the indices along its dimension
            # so we can directly subtract the two dimensions to get a 2D tensor of (query_sequence x key_sequence)
            # with the distance between them
            distance = query_sequence - key_sequence

            if not (key_sequence.size <= self.max_position_embeddings):
                raise AssertionError(
                    f"key_sequence.size ({key_sequence.size}) exceeds "
                    f"max_position_embeddings ({self.max_position_embeddings})"
                )

            # we can then use that as an indirect index into the embedding table values to look up the features for that index
            # this is just a `gather` primitive op. The resulting tensor will
            # have all the dimensions of embedding_idx (query_sequence x key_sequence),
            # plus all the dimensions of `embed` that were not indirectly accessed (`embedding_range`).
            # this form of indirect indexing is more straightforward than either advanced indexing or torch.gather which both
            # have a lot of dependencies on the positions of indexing tensors.

            positional_embedding = self.distance_embedding.weight[
                self.max_position_embeddings - 1 + distance, features
            ]

            if self.position_embedding_type == "relative_key":
                # these were einsum ops in the positional code because they are not easy to fit to existing matmul operators
                # even though they are degenerate matmuls
                relative_position_scores = (q * positional_embedding).sum(features)
                attention_scores = attention_scores + relative_position_scores
            elif self.position_embedding_type == "relative_key_query":
                relative_position_scores_query = (q * positional_embedding).sum(
                    features
                )
                relative_position_scores_key = (k * positional_embedding).sum(features)
                attention_scores = (
                    attention_scores
                    + relative_position_scores_query
                    + relative_position_scores_key
                )

        attention_probs = attention_scores
        # Normalize the attention scores to probabilities.
        attention_probs = torch.softmax(attention_scores, dim=key_sequence)
        # # This is actually dropping out entire tokens to attend to, which might
        # # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = torch.nn.functional.dropout(
            attention_probs, p=self.dropout_prob
        )

        # similarly, we can replace the matmul with a direct listing of the outer product, which makes it clear
        # we are weighting the values v across all keys with the attention scores.
        context_layer = (attention_probs * v).sum(key_sequence)

        # finally, we convert back to a standard tensor by describing the layout of dimensions.
        # working in reverse to with_dims, the (heads, features) group flattens the dimensions into a single one.
        return context_layer.order(batch, query_sequence, [heads, features])



# Owner(s): ["module: dynamo"]

import functools
import itertools
from unittest import expectedFailure as xfail, skipIf as skipif, SkipTest

from pytest import raises as assert_raises

import torch._numpy as np
from torch._numpy.testing import (
    assert_,
    assert_allclose,
    assert_almost_equal,
    assert_array_equal,
    assert_equal,
    suppress_warnings,
)
from torch.testing._internal.common_utils import (
    instantiate_parametrized_tests,
    parametrize,
    run_tests,
    TestCase,
)


skip = functools.partial(skipif, True)


# Setup for optimize einsum
chars = "abcdefghij"
sizes = np.array([2, 3, 4, 5, 4, 3, 2, 6, 5, 4, 3])
global_size_dict = dict(zip(chars, sizes))


@instantiate_parametrized_tests
class TestEinsum(TestCase):
    def test_einsum_errors(self):
        for do_opt in [True, False]:
            # Need enough arguments
            assert_raises(
                (TypeError, IndexError, ValueError), np.einsum, optimize=do_opt
            )
            assert_raises((IndexError, ValueError), np.einsum, "", optimize=do_opt)

            # subscripts must be a string
            assert_raises((AttributeError, TypeError), np.einsum, 0, 0, optimize=do_opt)

            # out parameter must be an array
            assert_raises(TypeError, np.einsum, "", 0, out="test", optimize=do_opt)

            # order parameter must be a valid order
            assert_raises(
                (NotImplementedError, ValueError),
                np.einsum,
                "",
                0,
                order="W",
                optimize=do_opt,
            )

            # casting parameter must be a valid casting
            assert_raises(ValueError, np.einsum, "", 0, casting="blah", optimize=do_opt)

            # dtype parameter must be a valid dtype
            assert_raises(
                TypeError, np.einsum, "", 0, dtype="bad_data_type", optimize=do_opt
            )

            # other keyword arguments are rejected
            assert_raises(TypeError, np.einsum, "", 0, bad_arg=0, optimize=do_opt)

            # issue 4528 revealed a segfault with this call
            assert_raises(
                (RuntimeError, TypeError), np.einsum, *(None,) * 63, optimize=do_opt
            )

            # number of operands must match count in subscripts string
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "", 0, 0, optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, ",", 0, [0], [0], optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, ",", [0], optimize=do_opt
            )

            # can't have more subscripts than dimensions in the operand
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "i", 0, optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "ij", [0, 0], optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "...i", 0, optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "i...j", [0, 0], optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "i...", 0, optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "ij...", [0, 0], optimize=do_opt
            )

            # invalid ellipsis
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "i..", [0, 0], optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, ".i...", [0, 0], optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "j->..j", [0, 0], optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError),
                np.einsum,
                "j->.j...",
                [0, 0],
                optimize=do_opt,
            )

            # invalid subscript character
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "i%...", [0, 0], optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "...j$", [0, 0], optimize=do_opt
            )
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "i->&", [0, 0], optimize=do_opt
            )

            # output subscripts must appear in input
            assert_raises(
                (RuntimeError, ValueError), np.einsum, "i->ij", [0, 0], optimize=do_opt
            )

            # output subscripts may only be specified once
            assert_raises(
                (RuntimeError, ValueError),
                np.einsum,
                "ij->jij",
                [[0, 0], [0, 0]],
                optimize=do_opt,
            )

            # dimensions much match when being collapsed
            assert_raises(
                (RuntimeError, ValueError),
                np.einsum,
                "ii",
                np.arange(6).reshape(2, 3),
                optimize=do_opt,
            )
            assert_raises(
                (RuntimeError, ValueError),
                np.einsum,
                "ii->i",
                np.arange(6).reshape(2, 3),
                optimize=do_opt,
            )

            # broadcasting to new dimensions must be enabled explicitly
            assert_raises(
                (RuntimeError, ValueError),
                np.einsum,
                "i",
                np.arange(6).reshape(2, 3),
                optimize=do_opt,
            )
            assert_raises(
                (RuntimeError, ValueError),
                np.einsum,
                "i->i",
                [[0, 1], [0, 1]],
                out=np.arange(4).reshape(2, 2),
                optimize=do_opt,
            )
            with assert_raises((RuntimeError, ValueError)):  # , match="'b'"):
                # gh-11221 - 'c' erroneously appeared in the error message
                a = np.ones((3, 3, 4, 5, 6))
                b = np.ones((3, 4, 5))
                np.einsum("aabcb,abc", a, b)

            # Check order kwarg, asanyarray allows 1d to pass through
            assert_raises(
                (NotImplementedError, ValueError),
                np.einsum,
                "i->i",
                np.arange(6).reshape(-1, 1),
                optimize=do_opt,
                order="d",
            )

    @xfail  # (reason="a view into smth else")
    def test_einsum_views(self):
        # pass-through
        for do_opt in [True, False]:
            a = np.arange(6)
            a = a.reshape(2, 3)

            b = np.einsum("...", a, optimize=do_opt)
            assert_(b.tensor._base is a.tensor)

            b = np.einsum(a, [Ellipsis], optimize=do_opt)
            assert_(b.base is a)

            b = np.einsum("ij", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, a)

            b = np.einsum(a, [0, 1], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, a)

            # output is writeable whenever input is writeable
            b = np.einsum("...", a, optimize=do_opt)
            assert_(b.flags["WRITEABLE"])
            a.flags["WRITEABLE"] = False
            b = np.einsum("...", a, optimize=do_opt)
            assert_(not b.flags["WRITEABLE"])

            # transpose
            a = np.arange(6)
            a.shape = (2, 3)

            b = np.einsum("ji", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, a.T)

            b = np.einsum(a, [1, 0], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, a.T)

            # diagonal
            a = np.arange(9)
            a.shape = (3, 3)

            b = np.einsum("ii->i", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a[i, i] for i in range(3)])

            b = np.einsum(a, [0, 0], [0], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a[i, i] for i in range(3)])

            # diagonal with various ways of broadcasting an additional dimension
            a = np.arange(27)
            a.shape = (3, 3, 3)

            b = np.einsum("...ii->...i", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [[x[i, i] for i in range(3)] for x in a])

            b = np.einsum(a, [Ellipsis, 0, 0], [Ellipsis, 0], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [[x[i, i] for i in range(3)] for x in a])

            b = np.einsum("ii...->...i", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [[x[i, i] for i in range(3)] for x in a.transpose(2, 0, 1)])

            b = np.einsum(a, [0, 0, Ellipsis], [Ellipsis, 0], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [[x[i, i] for i in range(3)] for x in a.transpose(2, 0, 1)])

            b = np.einsum("...ii->i...", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a[:, i, i] for i in range(3)])

            b = np.einsum(a, [Ellipsis, 0, 0], [0, Ellipsis], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a[:, i, i] for i in range(3)])

            b = np.einsum("jii->ij", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a[:, i, i] for i in range(3)])

            b = np.einsum(a, [1, 0, 0], [0, 1], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a[:, i, i] for i in range(3)])

            b = np.einsum("ii...->i...", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a.transpose(2, 0, 1)[:, i, i] for i in range(3)])

            b = np.einsum(a, [0, 0, Ellipsis], [0, Ellipsis], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a.transpose(2, 0, 1)[:, i, i] for i in range(3)])

            b = np.einsum("i...i->i...", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a.transpose(1, 0, 2)[:, i, i] for i in range(3)])

            b = np.einsum(a, [0, Ellipsis, 0], [0, Ellipsis], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a.transpose(1, 0, 2)[:, i, i] for i in range(3)])

            b = np.einsum("i...i->...i", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [[x[i, i] for i in range(3)] for x in a.transpose(1, 0, 2)])

            b = np.einsum(a, [0, Ellipsis, 0], [Ellipsis, 0], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [[x[i, i] for i in range(3)] for x in a.transpose(1, 0, 2)])

            # triple diagonal
            a = np.arange(27)
            a.shape = (3, 3, 3)

            b = np.einsum("iii->i", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a[i, i, i] for i in range(3)])

            b = np.einsum(a, [0, 0, 0], [0], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, [a[i, i, i] for i in range(3)])

            # swap axes
            a = np.arange(24)
            a.shape = (2, 3, 4)

            b = np.einsum("ijk->jik", a, optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, a.swapaxes(0, 1))

            b = np.einsum(a, [0, 1, 2], [1, 0, 2], optimize=do_opt)
            assert_(b.base is a)
            assert_equal(b, a.swapaxes(0, 1))

    #  @np._no_nep50_warning()
    def check_einsum_sums(self, dtype, do_opt=False):
        dtype = np.dtype(dtype)
        # Check various sums.  Does many sizes to exercise unrolled loops.

        # sum(a, axis=-1)
        for n in range(1, 17):
            a = np.arange(n, dtype=dtype)
            assert_equal(
                np.einsum("i->", a, optimize=do_opt), np.sum(a, axis=-1).astype(dtype)
            )
            assert_equal(
                np.einsum(a, [0], [], optimize=do_opt), np.sum(a, axis=-1).astype(dtype)
            )

        for n in range(1, 17):
            a = np.arange(2 * 3 * n, dtype=dtype).reshape(2, 3, n)
            assert_equal(
                np.einsum("...i->...", a, optimize=do_opt),
                np.sum(a, axis=-1).astype(dtype),
            )
            assert_equal(
                np.einsum(a, [Ellipsis, 0], [Ellipsis], optimize=do_opt),
                np.sum(a, axis=-1).astype(dtype),
            )

        # sum(a, axis=0)
        for n in range(1, 17):
            a = np.arange(2 * n, dtype=dtype).reshape(2, n)
            assert_equal(
                np.einsum("i...->...", a, optimize=do_opt),
                np.sum(a, axis=0).astype(dtype),
            )
            assert_equal(
                np.einsum(a, [0, Ellipsis], [Ellipsis], optimize=do_opt),
                np.sum(a, axis=0).astype(dtype),
            )

        for n in range(1, 17):
            a = np.arange(2 * 3 * n, dtype=dtype).reshape(2, 3, n)
            assert_equal(
                np.einsum("i...->...", a, optimize=do_opt),
                np.sum(a, axis=0).astype(dtype),
            )
            assert_equal(
                np.einsum(a, [0, Ellipsis], [Ellipsis], optimize=do_opt),
                np.sum(a, axis=0).astype(dtype),
            )

        # trace(a)
        for n in range(1, 17):
            a = np.arange(n * n, dtype=dtype).reshape(n, n)
            assert_equal(np.einsum("ii", a, optimize=do_opt), np.trace(a).astype(dtype))
            assert_equal(
                np.einsum(a, [0, 0], optimize=do_opt),  # torch?
                np.trace(a).astype(dtype),
            )

            # gh-15961: should accept numpy int64 type in subscript list
        #     np_array = np.asarray([0, 0])
        #     assert_equal(np.einsum(a, np_array, optimize=do_opt),
        #                  np.trace(a).astype(dtype))
        #     assert_equal(np.einsum(a, list(np_array), optimize=do_opt),
        #                  np.trace(a).astype(dtype))

        # multiply(a, b)
        assert_equal(np.einsum("..., ...", 3, 4), 12)  # scalar case
        for n in range(1, 17):
            a = np.arange(3 * n, dtype=dtype).reshape(3, n)
            b = np.arange(2 * 3 * n, dtype=dtype).reshape(2, 3, n)
            assert_equal(
                np.einsum("..., ...", a, b, optimize=do_opt), np.multiply(a, b)
            )
            assert_equal(
                np.einsum(a, [Ellipsis], b, [Ellipsis], optimize=do_opt),
                np.multiply(a, b),
            )

        # inner(a,b)
        for n in range(1, 17):
            a = np.arange(2 * 3 * n, dtype=dtype).reshape(2, 3, n)
            b = np.arange(n, dtype=dtype)
            assert_equal(np.einsum("...i, ...i", a, b, optimize=do_opt), np.inner(a, b))
            assert_equal(
                np.einsum(a, [Ellipsis, 0], b, [Ellipsis, 0], optimize=do_opt),
                np.inner(a, b),
            )

        for n in range(1, 11):
            a = np.arange(n * 3 * 2, dtype=dtype).reshape(n, 3, 2)
            b = np.arange(n, dtype=dtype)
            assert_equal(
                np.einsum("i..., i...", a, b, optimize=do_opt), np.inner(a.T, b.T).T
            )
            assert_equal(
                np.einsum(a, [0, Ellipsis], b, [0, Ellipsis], optimize=do_opt),
                np.inner(a.T, b.T).T,
            )

        # outer(a,b)
        for n in range(1, 17):
            a = np.arange(3, dtype=dtype) + 1
            b = np.arange(n, dtype=dtype) + 1
            assert_equal(np.einsum("i,j", a, b, optimize=do_opt), np.outer(a, b))
            assert_equal(np.einsum(a, [0], b, [1], optimize=do_opt), np.outer(a, b))

        # Suppress the complex warnings for the 'as f8' tests
        with suppress_warnings():
            # matvec(a,b) / a.dot(b) where a is matrix, b is vector
            for n in range(1, 17):
                a = np.arange(4 * n, dtype=dtype).reshape(4, n)
                b = np.arange(n, dtype=dtype)
                assert_equal(np.einsum("ij, j", a, b, optimize=do_opt), np.dot(a, b))
                assert_equal(
                    np.einsum(a, [0, 1], b, [1], optimize=do_opt), np.dot(a, b)
                )

                c = np.arange(4, dtype=dtype)
                np.einsum(
                    "ij,j", a, b, out=c, dtype="f8", casting="unsafe", optimize=do_opt
                )
                assert_equal(c, np.dot(a.astype("f8"), b.astype("f8")).astype(dtype))
                c[...] = 0
                np.einsum(
                    a,
                    [0, 1],
                    b,
                    [1],
                    out=c,
                    dtype="f8",
                    casting="unsafe",
                    optimize=do_opt,
                )
                assert_equal(c, np.dot(a.astype("f8"), b.astype("f8")).astype(dtype))

            for n in range(1, 17):
                a = np.arange(4 * n, dtype=dtype).reshape(4, n)
                b = np.arange(n, dtype=dtype)
                assert_equal(
                    np.einsum("ji,j", a.T, b.T, optimize=do_opt), np.dot(b.T, a.T)
                )
                assert_equal(
                    np.einsum(a.T, [1, 0], b.T, [1], optimize=do_opt), np.dot(b.T, a.T)
                )

                c = np.arange(4, dtype=dtype)
                np.einsum(
                    "ji,j",
                    a.T,
                    b.T,
                    out=c,
                    dtype="f8",
                    casting="unsafe",
                    optimize=do_opt,
                )
                assert_equal(
                    c, np.dot(b.T.astype("f8"), a.T.astype("f8")).astype(dtype)
                )
                c[...] = 0
                np.einsum(
                    a.T,
                    [1, 0],
                    b.T,
                    [1],
                    out=c,
                    dtype="f8",
                    casting="unsafe",
                    optimize=do_opt,
                )
                assert_equal(
                    c, np.dot(b.T.astype("f8"), a.T.astype("f8")).astype(dtype)
                )

            # matmat(a,b) / a.dot(b) where a is matrix, b is matrix
            for n in range(1, 17):
                if n < 8 or dtype != "f2":
                    a = np.arange(4 * n, dtype=dtype).reshape(4, n)
                    b = np.arange(n * 6, dtype=dtype).reshape(n, 6)
                    assert_equal(
                        np.einsum("ij,jk", a, b, optimize=do_opt), np.dot(a, b)
                    )
                    assert_equal(
                        np.einsum(a, [0, 1], b, [1, 2], optimize=do_opt), np.dot(a, b)
                    )

            for n in range(1, 17):
                a = np.arange(4 * n, dtype=dtype).reshape(4, n)
                b = np.arange(n * 6, dtype=dtype).reshape(n, 6)
                c = np.arange(24, dtype=dtype).reshape(4, 6)
                np.einsum(
                    "ij,jk", a, b, out=c, dtype="f8", casting="unsafe", optimize=do_opt
                )
                assert_equal(c, np.dot(a.astype("f8"), b.astype("f8")).astype(dtype))
                c[...] = 0
                np.einsum(
                    a,
                    [0, 1],
                    b,
                    [1, 2],
                    out=c,
                    dtype="f8",
                    casting="unsafe",
                    optimize=do_opt,
                )
                assert_equal(c, np.dot(a.astype("f8"), b.astype("f8")).astype(dtype))

            # matrix triple product (note this is not currently an efficient
            # way to multiply 3 matrices)
            a = np.arange(12, dtype=dtype).reshape(3, 4)
            b = np.arange(20, dtype=dtype).reshape(4, 5)
            c = np.arange(30, dtype=dtype).reshape(5, 6)
            if dtype != "f2":
                assert_equal(
                    np.einsum("ij,jk,kl", a, b, c, optimize=do_opt), a.dot(b).dot(c)
                )
                assert_equal(
                    np.einsum(a, [0, 1], b, [1, 2], c, [2, 3], optimize=do_opt),
                    a.dot(b).dot(c),
                )

            d = np.arange(18, dtype=dtype).reshape(3, 6)
            np.einsum(
                "ij,jk,kl",
                a,
                b,
                c,
                out=d,
                dtype="f8",
                casting="unsafe",
                optimize=do_opt,
            )
            tgt = a.astype("f8").dot(b.astype("f8"))
            tgt = tgt.dot(c.astype("f8")).astype(dtype)
            assert_equal(d, tgt)

            d[...] = 0
            np.einsum(
                a,
                [0, 1],
                b,
                [1, 2],
                c,
                [2, 3],
                out=d,
                dtype="f8",
                casting="unsafe",
                optimize=do_opt,
            )
            tgt = a.astype("f8").dot(b.astype("f8"))
            tgt = tgt.dot(c.astype("f8")).astype(dtype)
            assert_equal(d, tgt)

            # tensordot(a, b)
            if np.dtype(dtype) != np.dtype("f2"):
                a = np.arange(60, dtype=dtype).reshape(3, 4, 5)
                b = np.arange(24, dtype=dtype).reshape(4, 3, 2)
                assert_equal(
                    np.einsum("ijk, jil -> kl", a, b),
                    np.tensordot(a, b, axes=([1, 0], [0, 1])),
                )
                assert_equal(
                    np.einsum(a, [0, 1, 2], b, [1, 0, 3], [2, 3]),
                    np.tensordot(a, b, axes=([1, 0], [0, 1])),
                )

                c = np.arange(10, dtype=dtype).reshape(5, 2)
                np.einsum(
                    "ijk,jil->kl",
                    a,
                    b,
                    out=c,
                    dtype="f8",
                    casting="unsafe",
                    optimize=do_opt,
                )
                assert_equal(
                    c,
                    np.tensordot(
                        a.astype("f8"), b.astype("f8"), axes=([1, 0], [0, 1])
                    ).astype(dtype),
                )
                c[...] = 0
                np.einsum(
                    a,
                    [0, 1, 2],
                    b,
                    [1, 0, 3],
                    [2, 3],
                    out=c,
                    dtype="f8",
                    casting="unsafe",
                    optimize=do_opt,
                )
                assert_equal(
                    c,
                    np.tensordot(
                        a.astype("f8"), b.astype("f8"), axes=([1, 0], [0, 1])
                    ).astype(dtype),
                )

        # logical_and(logical_and(a!=0, b!=0), c!=0)
        neg_val = -2 if dtype.kind != "u" else np.iinfo(dtype).max - 1
        a = np.array([1, 3, neg_val, 0, 12, 13, 0, 1], dtype=dtype)
        b = np.array([0, 3.5, 0.0, neg_val, 0, 1, 3, 12], dtype=dtype)
        c = np.array([True, True, False, True, True, False, True, True])

        assert_equal(
            np.einsum(
                "i,i,i->i", a, b, c, dtype="?", casting="unsafe", optimize=do_opt
            ),
            np.logical_and(np.logical_and(a != 0, b != 0), c != 0),
        )
        assert_equal(
            np.einsum(a, [0], b, [0], c, [0], [0], dtype="?", casting="unsafe"),
            np.logical_and(np.logical_and(a != 0, b != 0), c != 0),
        )

        a = np.arange(9, dtype=dtype)
        assert_equal(np.einsum(",i->", 3, a), 3 * np.sum(a))
        assert_equal(np.einsum(3, [], a, [0], []), 3 * np.sum(a))
        assert_equal(np.einsum("i,->", a, 3), 3 * np.sum(a))
        assert_equal(np.einsum(a, [0], 3, [], []), 3 * np.sum(a))

        # Various stride0, contiguous, and SSE aligned variants
        for n in range(1, 25):
            a = np.arange(n, dtype=dtype)
            if np.dtype(dtype).itemsize > 1:
                assert_equal(
                    np.einsum("...,...", a, a, optimize=do_opt), np.multiply(a, a)
                )
                assert_equal(np.einsum("i,i", a, a, optimize=do_opt), np.dot(a, a))
                assert_equal(np.einsum("i,->i", a, 2, optimize=do_opt), 2 * a)
                assert_equal(np.einsum(",i->i", 2, a, optimize=do_opt), 2 * a)
                assert_equal(np.einsum("i,->", a, 2, optimize=do_opt), 2 * np.sum(a))
                assert_equal(np.einsum(",i->", 2, a, optimize=do_opt), 2 * np.sum(a))

                assert_equal(
                    np.einsum("...,...", a[1:], a[:-1], optimize=do_opt),
                    np.multiply(a[1:], a[:-1]),
                )
                assert_equal(
                    np.einsum("i,i", a[1:], a[:-1], optimize=do_opt),
                    np.dot(a[1:], a[:-1]),
                )
                assert_equal(np.einsum("i,->i", a[1:], 2, optimize=do_opt), 2 * a[1:])
                assert_equal(np.einsum(",i->i", 2, a[1:], optimize=do_opt), 2 * a[1:])
                assert_equal(
                    np.einsum("i,->", a[1:], 2, optimize=do_opt), 2 * np.sum(a[1:])
                )
                assert_equal(
                    np.einsum(",i->", 2, a[1:], optimize=do_opt), 2 * np.sum(a[1:])
                )

        # An object array, summed as the data type
        #    a = np.arange(9, dtype=object)
        #
        #    b = np.einsum("i->", a, dtype=dtype, casting='unsafe')
        #    assert_equal(b, np.sum(a))
        #    assert_equal(b.dtype, np.dtype(dtype))
        #
        #    b = np.einsum(a, [0], [], dtype=dtype, casting='unsafe')
        #    assert_equal(b, np.sum(a))
        #    assert_equal(b.dtype, np.dtype(dtype))

        # A case which was failing (ticket #1885)
        p = np.arange(2) + 1
        q = np.arange(4).reshape(2, 2) + 3
        r = np.arange(4).reshape(2, 2) + 7
        assert_equal(np.einsum("z,mz,zm->", p, q, r), 253)

        # singleton dimensions broadcast (gh-10343)
        p = np.ones((10, 2))
        q = np.ones((1, 2))
        assert_array_equal(
            np.einsum("ij,ij->j", p, q, optimize=True),
            np.einsum("ij,ij->j", p, q, optimize=False),
        )
        assert_array_equal(np.einsum("ij,ij->j", p, q, optimize=True), [10.0] * 2)

        # a blas-compatible contraction broadcasting case which was failing
        # for optimize=True (ticket #10930)
        x = np.array([2.0, 3.0])
        y = np.array([4.0])
        assert_array_equal(np.einsum("i, i", x, y, optimize=False), 20.0)
        assert_array_equal(np.einsum("i, i", x, y, optimize=True), 20.0)

        # all-ones array was bypassing bug (ticket #10930)
        p = np.ones((1, 5)) / 2
        q = np.ones((5, 5)) / 2
        for optimize in (True, False):
            assert_array_equal(
                np.einsum("...ij,...jk->...ik", p, p, optimize=optimize),
                np.einsum("...ij,...jk->...ik", p, q, optimize=optimize),
            )
            assert_array_equal(
                np.einsum("...ij,...jk->...ik", p, q, optimize=optimize),
                np.full((1, 5), 1.25),
            )

        # Cases which were failing (gh-10899)
        x = np.eye(2, dtype=dtype)
        y = np.ones(2, dtype=dtype)
        assert_array_equal(
            np.einsum("ji,i->", x, y, optimize=optimize), [2.0]
        )  # contig_contig_outstride0_two
        assert_array_equal(
            np.einsum("i,ij->", y, x, optimize=optimize), [2.0]
        )  # stride0_contig_outstride0_two
        assert_array_equal(
            np.einsum("ij,i->", x, y, optimize=optimize), [2.0]
        )  # contig_stride0_outstride0_two

    @xfail  # (reason="int overflow differs in numpy and pytorch")
    def test_einsum_sums_int8(self):
        self.check_einsum_sums("i1")

    @xfail  # (reason="int overflow differs in numpy and pytorch")
    def test_einsum_sums_uint8(self):
        self.check_einsum_sums("u1")

    @xfail  # (reason="int overflow differs in numpy and pytorch")
    def test_einsum_sums_int16(self):
        self.check_einsum_sums("i2")

    def test_einsum_sums_int32(self):
        self.check_einsum_sums("i4")
        self.check_einsum_sums("i4", True)

    def test_einsum_sums_int64(self):
        self.check_einsum_sums("i8")

    @xfail  # (reason="np.float16(4641) == 4640.0")
    def test_einsum_sums_float16(self):
        self.check_einsum_sums("f2")

    def test_einsum_sums_float32(self):
        self.check_einsum_sums("f4")

    def test_einsum_sums_float64(self):
        self.check_einsum_sums("f8")
        self.check_einsum_sums("f8", True)

    def test_einsum_sums_cfloat64(self):
        self.check_einsum_sums("c8")
        self.check_einsum_sums("c8", True)

    def test_einsum_sums_cfloat128(self):
        self.check_einsum_sums("c16")

    def test_einsum_misc(self):
        # This call used to crash because of a bug in
        # PyArray_AssignZero
        a = np.ones((1, 2))
        b = np.ones((2, 2, 1))
        assert_equal(np.einsum("ij...,j...->i...", a, b), [[[2], [2]]])
        assert_equal(np.einsum("ij...,j...->i...", a, b, optimize=True), [[[2], [2]]])

        # Regression test for issue #10369 (test unicode inputs with Python 2)
        assert_equal(np.einsum("ij...,j...->i...", a, b), [[[2], [2]]])
        assert_equal(np.einsum("...i,...i", [1, 2, 3], [2, 3, 4]), 20)
        assert_equal(
            np.einsum("...i,...i", [1, 2, 3], [2, 3, 4], optimize="greedy"), 20
        )

        # The iterator had an issue with buffering this reduction
        a = np.ones((5, 12, 4, 2, 3), np.int64)
        b = np.ones((5, 12, 11), np.int64)
        assert_equal(
            np.einsum("ijklm,ijn,ijn->", a, b, b), np.einsum("ijklm,ijn->", a, b)
        )
        assert_equal(
            np.einsum("ijklm,ijn,ijn->", a, b, b, optimize=True),
            np.einsum("ijklm,ijn->", a, b, optimize=True),
        )

        # Issue #2027, was a problem in the contiguous 3-argument
        # inner loop implementation
        a = np.arange(1, 3)
        b = np.arange(1, 5).reshape(2, 2)
        c = np.arange(1, 9).reshape(4, 2)
        assert_equal(
            np.einsum("x,yx,zx->xzy", a, b, c),
            [
                [[1, 3], [3, 9], [5, 15], [7, 21]],
                [[8, 16], [16, 32], [24, 48], [32, 64]],
            ],
        )
        assert_equal(
            np.einsum("x,yx,zx->xzy", a, b, c, optimize=True),
            [
                [[1, 3], [3, 9], [5, 15], [7, 21]],
                [[8, 16], [16, 32], [24, 48], [32, 64]],
            ],
        )

        # Ensure explicitly setting out=None does not cause an error
        # see issue gh-15776 and issue gh-15256
        assert_equal(np.einsum("i,j", [1], [2], out=None), [[2]])

    def test_subscript_range(self):
        # Issue #7741, make sure that all letters of Latin alphabet (both uppercase & lowercase) can be used
        # when creating a subscript from arrays
        a = np.ones((2, 3))
        b = np.ones((3, 4))
        np.einsum(a, [0, 20], b, [20, 2], [0, 2], optimize=False)
        np.einsum(a, [0, 27], b, [27, 2], [0, 2], optimize=False)
        np.einsum(a, [0, 51], b, [51, 2], [0, 2], optimize=False)
        assert_raises(
            ValueError,
            lambda: np.einsum(a, [0, 52], b, [52, 2], [0, 2], optimize=False),
        )
        assert_raises(
            ValueError,
            lambda: np.einsum(a, [-1, 5], b, [5, 2], [-1, 2], optimize=False),
        )

    def test_einsum_broadcast(self):
        # Issue #2455 change in handling ellipsis
        # remove the 'middle broadcast' error
        # only use the 'RIGHT' iteration in prepare_op_axes
        # adds auto broadcast on left where it belongs
        # broadcast on right has to be explicit
        # We need to test the optimized parsing as well

        A = np.arange(2 * 3 * 4).reshape(2, 3, 4)
        B = np.arange(3)
        ref = np.einsum("ijk,j->ijk", A, B, optimize=False)
        for opt in [True, False]:
            assert_equal(np.einsum("ij...,j...->ij...", A, B, optimize=opt), ref)
            assert_equal(np.einsum("ij...,...j->ij...", A, B, optimize=opt), ref)
            assert_equal(
                np.einsum("ij...,j->ij...", A, B, optimize=opt), ref
            )  # used to raise error

        A = np.arange(12).reshape((4, 3))
        B = np.arange(6).reshape((3, 2))
        ref = np.einsum("ik,kj->ij", A, B, optimize=False)
        for opt in [True, False]:
            assert_equal(np.einsum("ik...,k...->i...", A, B, optimize=opt), ref)
            assert_equal(np.einsum("ik...,...kj->i...j", A, B, optimize=opt), ref)
            assert_equal(
                np.einsum("...k,kj", A, B, optimize=opt), ref
            )  # used to raise error
            assert_equal(
                np.einsum("ik,k...->i...", A, B, optimize=opt), ref
            )  # used to raise error

        dims = [2, 3, 4, 5]
        a = np.arange(np.prod(dims)).reshape(dims)
        v = np.arange(dims[2])
        ref = np.einsum("ijkl,k->ijl", a, v, optimize=False)
        for opt in [True, False]:
            assert_equal(np.einsum("ijkl,k", a, v, optimize=opt), ref)
            assert_equal(
                np.einsum("...kl,k", a, v, optimize=opt), ref
            )  # used to raise error
            assert_equal(np.einsum("...kl,k...", a, v, optimize=opt), ref)

        J, K, M = 160, 160, 120
        A = np.arange(J * K * M).reshape(1, 1, 1, J, K, M)
        B = np.arange(J * K * M * 3).reshape(J, K, M, 3)
        ref = np.einsum("...lmn,...lmno->...o", A, B, optimize=False)
        for opt in [True, False]:
            assert_equal(
                np.einsum("...lmn,lmno->...o", A, B, optimize=opt), ref
            )  # used to raise error

    def test_einsum_fixedstridebug(self):
        # Issue #4485 obscure einsum bug
        # This case revealed a bug in nditer where it reported a stride
        # as 'fixed' (0) when it was in fact not fixed during processing
        # (0 or 4). The reason for the bug was that the check for a fixed
        # stride was using the information from the 2D inner loop reuse
        # to restrict the iteration dimensions it had to validate to be
        # the same, but that 2D inner loop reuse logic is only triggered
        # during the buffer copying step, and hence it was invalid to
        # rely on those values. The fix is to check all the dimensions
        # of the stride in question, which in the test case reveals that
        # the stride is not fixed.
        #
        # NOTE: This test is triggered by the fact that the default buffersize,
        #       used by einsum, is 8192, and 3*2731 = 8193, is larger than that
        #       and results in a mismatch between the buffering and the
        #       striding for operand A.
        A = np.arange(2 * 3).reshape(2, 3).astype(np.float32)
        B = np.arange(2 * 3 * 2731).reshape(2, 3, 2731).astype(np.int16)
        es = np.einsum("cl, cpx->lpx", A, B)
        tp = np.tensordot(A, B, axes=(0, 0))
        assert_equal(es, tp)
        # The following is the original test case from the bug report,
        # made repeatable by changing random arrays to aranges.  # codespell:ignore aranges
        A = np.arange(3 * 3).reshape(3, 3).astype(np.float64)
        B = np.arange(3 * 3 * 64 * 64).reshape(3, 3, 64, 64).astype(np.float32)
        es = np.einsum("cl, cpxy->lpxy", A, B)
        tp = np.tensordot(A, B, axes=(0, 0))
        assert_equal(es, tp)

    def test_einsum_fixed_collapsingbug(self):
        # Issue #5147.
        # The bug only occurred when output argument of einssum was used.
        x = np.random.normal(0, 1, (5, 5, 5, 5))
        y1 = np.zeros((5, 5))
        np.einsum("aabb->ab", x, out=y1)
        idx = np.arange(5)
        y2 = x[idx[:, None], idx[:, None], idx, idx]
        assert_equal(y1, y2)

    def test_einsum_failed_on_p9_and_s390x(self):
        # Issues gh-14692 and gh-12689
        # Bug with signed vs unsigned char errored on power9 and s390x Linux
        tensor = np.random.random_sample((10, 10, 10, 10))
        x = np.einsum("ijij->", tensor)
        y = tensor.trace(axis1=0, axis2=2).trace()
        assert_allclose(x, y)

    @xfail  # (reason="no base")
    def test_einsum_all_contig_non_contig_output(self):
        # Issue gh-5907, tests that the all contiguous special case
        # actually checks the contiguity of the output
        x = np.ones((5, 5))
        out = np.ones(10)[::2]
        correct_base = np.ones(10)
        correct_base[::2] = 5
        # Always worked (inner iteration is done with 0-stride):
        np.einsum("mi,mi,mi->m", x, x, x, out=out)
        assert_array_equal(out.base, correct_base)
        # Example 1:
        out = np.ones(10)[::2]
        np.einsum("im,im,im->m", x, x, x, out=out)
        assert_array_equal(out.base, correct_base)
        # Example 2, buffering causes x to be contiguous but
        # special cases do not catch the operation before:
        out = np.ones((2, 2, 2))[..., 0]
        correct_base = np.ones((2, 2, 2))
        correct_base[..., 0] = 2
        x = np.ones((2, 2), np.float32)
        np.einsum("ij,jk->ik", x, x, out=out)
        assert_array_equal(out.base, correct_base)

    @parametrize("dtype", np.typecodes["AllFloat"] + np.typecodes["AllInteger"])
    def test_different_paths(self, dtype):
        # Test originally added to cover broken float16 path: gh-20305
        # Likely most are covered elsewhere, at least partially.
        dtype = np.dtype(dtype)
        # Simple test, designed to exercise most specialized code paths,
        # note the +0.5 for floats.  This makes sure we use a float value
        # where the results must be exact.
        arr = (np.arange(7) + 0.5).astype(dtype)
        scalar = np.array(2, dtype=dtype)

        # contig -> scalar:
        res = np.einsum("i->", arr)
        if res != arr.sum():
            raise AssertionError(f"Expected res == arr.sum(), got {res} vs {arr.sum()}")
        # contig, contig -> contig:
        res = np.einsum("i,i->i", arr, arr)
        assert_array_equal(res, arr * arr)
        # noncontig, noncontig -> contig:
        res = np.einsum("i,i->i", arr.repeat(2)[::2], arr.repeat(2)[::2])
        assert_array_equal(res, arr * arr)
        # contig + contig -> scalar
        einsum_result = np.einsum("i,i->", arr, arr)
        expected = (arr * arr).sum()
        if einsum_result != expected:
            raise AssertionError(
                f"Expected einsum result == {expected}, got {einsum_result}"
            )
        # contig + scalar -> contig (with out)
        out = np.ones(7, dtype=dtype)
        res = np.einsum("i,->i", arr, dtype.type(2), out=out)
        assert_array_equal(res, arr * dtype.type(2))
        # scalar + contig -> contig (with out)
        res = np.einsum(",i->i", scalar, arr)
        assert_array_equal(res, arr * dtype.type(2))
        # scalar + contig -> scalar
        res = np.einsum(",i->", scalar, arr)
        # Use einsum to compare to not have difference due to sum round-offs:
        expected = np.einsum("i->", scalar * arr)
        if res != expected:
            raise AssertionError(f"Expected res == {expected}, got {res}")
        # contig + scalar -> scalar
        res = np.einsum("i,->", arr, scalar)
        # Use einsum to compare to not have difference due to sum round-offs:
        expected = np.einsum("i->", scalar * arr)
        if res != expected:
            raise AssertionError(f"Expected res == {expected}, got {res}")
        # contig + contig + contig -> scalar

        if dtype in ["e", "B", "b"]:
            # FIXME make xfail
            raise SkipTest("overflow differs in pytorch and numpy")

        arr = np.array([0.5, 0.5, 0.25, 4.5, 3.0], dtype=dtype)
        res = np.einsum("i,i,i->", arr, arr, arr)
        assert_array_equal(res, (arr * arr * arr).sum())
        # four arrays:
        res = np.einsum("i,i,i,i->", arr, arr, arr, arr)
        assert_array_equal(res, (arr * arr * arr * arr).sum())

    def test_small_boolean_arrays(self):
        # See gh-5946.
        # Use array of True embedded in False.
        a = np.zeros((16, 1, 1), dtype=np.bool_)[:2]
        a[...] = True
        out = np.zeros((16, 1, 1), dtype=np.bool_)[:2]
        tgt = np.ones((2, 1, 1), dtype=np.bool_)
        res = np.einsum("...ij,...jk->...ik", a, a, out=out)
        assert_equal(res, tgt)

    def test_out_is_res(self):
        a = np.arange(9).reshape(3, 3)
        res = np.einsum("...ij,...jk->...ik", a, a, out=a)
        if res is not a:
            raise AssertionError("Expected res is a")

    def optimize_compare(self, subscripts, operands=None):
        # Tests all paths of the optimization function against
        # conventional einsum
        if operands is None:
            args = [subscripts]
            terms = subscripts.split("->")[0].split(",")
            for term in terms:
                dims = [global_size_dict[x] for x in term]
                args.append(np.random.rand(*dims))
        else:
            args = [subscripts] + operands

        noopt = np.einsum(*args, optimize=False)
        opt = np.einsum(*args, optimize="greedy")
        assert_almost_equal(opt, noopt)
        opt = np.einsum(*args, optimize="optimal")
        assert_almost_equal(opt, noopt)

    def test_hadamard_like_products(self):
        # Hadamard outer products
        self.optimize_compare("a,ab,abc->abc")
        self.optimize_compare("a,b,ab->ab")

    def test_index_transformations(self):
        # Simple index transformation cases
        self.optimize_compare("ea,fb,gc,hd,abcd->efgh")
        self.optimize_compare("ea,fb,abcd,gc,hd->efgh")
        self.optimize_compare("abcd,ea,fb,gc,hd->efgh")

    def test_complex(self):
        # Long test cases
        self.optimize_compare("acdf,jbje,gihb,hfac,gfac,gifabc,hfac")
        self.optimize_compare("acdf,jbje,gihb,hfac,gfac,gifabc,hfac")
        self.optimize_compare("cd,bdhe,aidb,hgca,gc,hgibcd,hgac")
        self.optimize_compare("abhe,hidj,jgba,hiab,gab")
        self.optimize_compare("bde,cdh,agdb,hica,ibd,hgicd,hiac")
        self.optimize_compare("chd,bde,agbc,hiad,hgc,hgi,hiad")
        self.optimize_compare("chd,bde,agbc,hiad,bdi,cgh,agdb")
        self.optimize_compare("bdhe,acad,hiab,agac,hibd")

    def test_collapse(self):
        # Inner products
        self.optimize_compare("ab,ab,c->")
        self.optimize_compare("ab,ab,c->c")
        self.optimize_compare("ab,ab,cd,cd->")
        self.optimize_compare("ab,ab,cd,cd->ac")
        self.optimize_compare("ab,ab,cd,cd->cd")
        self.optimize_compare("ab,ab,cd,cd,ef,ef->")

    def test_expand(self):
        # Outer products
        self.optimize_compare("ab,cd,ef->abcdef")
        self.optimize_compare("ab,cd,ef->acdf")
        self.optimize_compare("ab,cd,de->abcde")
        self.optimize_compare("ab,cd,de->be")
        self.optimize_compare("ab,bcd,cd->abcd")
        self.optimize_compare("ab,bcd,cd->abd")  # codespell:ignore

    def test_edge_cases(self):
        # Difficult edge cases for optimization
        self.optimize_compare("eb,cb,fb->cef")
        self.optimize_compare("dd,fb,be,cdb->cef")
        self.optimize_compare("bca,cdb,dbf,afc->")
        self.optimize_compare("dcc,fce,ea,dbf->ab")
        self.optimize_compare("fdf,cdd,ccd,afe->ae")
        self.optimize_compare("abcd,ad")
        self.optimize_compare("ed,fcd,ff,bcf->be")
        self.optimize_compare("baa,dcf,af,cde->be")
        self.optimize_compare("bd,db,eac->ace")
        self.optimize_compare("fff,fae,bef,def->abd")  # codespell:ignore
        self.optimize_compare("efc,dbc,acf,fd->abe")
        self.optimize_compare("ba,ac,da->bcd")

    def test_inner_product(self):
        # Inner products
        self.optimize_compare("ab,ab")
        self.optimize_compare("ab,ba")
        self.optimize_compare("abc,abc")
        self.optimize_compare("abc,bac")
        self.optimize_compare("abc,cba")

    def test_random_cases(self):
        # Randomly built test cases
        self.optimize_compare("aab,fa,df,ecc->bde")
        self.optimize_compare("ecb,fef,bad,ed->ac")
        self.optimize_compare("bcf,bbb,fbf,fc->")
        self.optimize_compare("bb,ff,be->e")
        self.optimize_compare("bcb,bb,fc,fff->")
        self.optimize_compare("fbb,dfd,fc,fc->")
        self.optimize_compare("afd,ba,cc,dc->bf")
        self.optimize_compare("adb,bc,fa,cfc->d")
        self.optimize_compare("bbd,bda,fc,db->acf")
        self.optimize_compare("dba,ead,cad->bce")
        self.optimize_compare("aef,fbc,dca->bde")

    def test_combined_views_mapping(self):
        # gh-10792
        a = np.arange(9).reshape(1, 1, 3, 1, 3)
        b = np.einsum("bbcdc->d", a)
        assert_equal(b, [12])

    def test_broadcasting_dot_cases(self):
        # Ensures broadcasting cases are not mistaken for GEMM

        a = np.random.rand(1, 5, 4)
        b = np.random.rand(4, 6)
        c = np.random.rand(5, 6)
        d = np.random.rand(10)

        self.optimize_compare("ijk,kl,jl", operands=[a, b, c])
        self.optimize_compare("ijk,kl,jl,i->i", operands=[a, b, c, d])

        e = np.random.rand(1, 1, 5, 4)
        f = np.random.rand(7, 7)
        self.optimize_compare("abjk,kl,jl", operands=[e, b, c])
        self.optimize_compare("abjk,kl,jl,ab->ab", operands=[e, b, c, f])

        # Edge case found in gh-11308
        g = np.arange(64).reshape(2, 4, 8)
        self.optimize_compare("obk,ijk->ioj", operands=[g, g])

    @xfail  # (reason="order='F' not supported")
    def test_output_order(self):
        # Ensure output order is respected for optimize cases, the below
        # contraction should yield a reshaped tensor view
        # gh-16415

        a = np.ones((2, 3, 5), order="F")
        b = np.ones((4, 3), order="F")

        for opt in [True, False]:
            tmp = np.einsum("...ft,mf->...mt", a, b, order="a", optimize=opt)
            assert_(tmp.flags.f_contiguous)

            tmp = np.einsum("...ft,mf->...mt", a, b, order="f", optimize=opt)
            assert_(tmp.flags.f_contiguous)

            tmp = np.einsum("...ft,mf->...mt", a, b, order="c", optimize=opt)
            assert_(tmp.flags.c_contiguous)

            tmp = np.einsum("...ft,mf->...mt", a, b, order="k", optimize=opt)
            assert_(tmp.flags.c_contiguous is False)
            assert_(tmp.flags.f_contiguous is False)

            tmp = np.einsum("...ft,mf->...mt", a, b, optimize=opt)
            assert_(tmp.flags.c_contiguous is False)
            assert_(tmp.flags.f_contiguous is False)

        c = np.ones((4, 3), order="C")
        for opt in [True, False]:
            tmp = np.einsum("...ft,mf->...mt", a, c, order="a", optimize=opt)
            assert_(tmp.flags.c_contiguous)

        d = np.ones((2, 3, 5), order="C")
        for opt in [True, False]:
            tmp = np.einsum("...ft,mf->...mt", d, c, order="a", optimize=opt)
            assert_(tmp.flags.c_contiguous)


@skip(reason="no pytorch analog")
class TestEinsumPath(TestCase):
    def build_operands(self, string, size_dict=global_size_dict):
        # Builds views based off initial operands
        operands = [string]
        terms = string.split("->")[0].split(",")
        for term in terms:
            dims = [size_dict[x] for x in term]
            operands.append(np.random.rand(*dims))

        return operands

    def assert_path_equal(self, comp, benchmark):
        # Checks if list of tuples are equivalent
        ret = len(comp) == len(benchmark)
        assert_(ret)
        for pos in range(len(comp) - 1):
            ret &= isinstance(comp[pos + 1], tuple)
            ret &= comp[pos + 1] == benchmark[pos + 1]
        assert_(ret)

    def test_memory_contraints(self):
        # Ensure memory constraints are satisfied

        outer_test = self.build_operands("a,b,c->abc")

        path, path_str = np.einsum_path(*outer_test, optimize=("greedy", 0))
        self.assert_path_equal(path, ["einsum_path", (0, 1, 2)])

        path, path_str = np.einsum_path(*outer_test, optimize=("optimal", 0))
        self.assert_path_equal(path, ["einsum_path", (0, 1, 2)])

        long_test = self.build_operands("acdf,jbje,gihb,hfac")
        path, path_str = np.einsum_path(*long_test, optimize=("greedy", 0))
        self.assert_path_equal(path, ["einsum_path", (0, 1, 2, 3)])

        path, path_str = np.einsum_path(*long_test, optimize=("optimal", 0))
        self.assert_path_equal(path, ["einsum_path", (0, 1, 2, 3)])

    def test_long_paths(self):
        # Long complex cases

        # Long test 1
        long_test1 = self.build_operands("acdf,jbje,gihb,hfac,gfac,gifabc,hfac")
        path, path_str = np.einsum_path(*long_test1, optimize="greedy")
        self.assert_path_equal(
            path, ["einsum_path", (3, 6), (3, 4), (2, 4), (2, 3), (0, 2), (0, 1)]
        )

        path, path_str = np.einsum_path(*long_test1, optimize="optimal")
        self.assert_path_equal(
            path, ["einsum_path", (3, 6), (3, 4), (2, 4), (2, 3), (0, 2), (0, 1)]
        )

        # Long test 2
        long_test2 = self.build_operands("chd,bde,agbc,hiad,bdi,cgh,agdb")
        path, path_str = np.einsum_path(*long_test2, optimize="greedy")
        self.assert_path_equal(
            path, ["einsum_path", (3, 4), (0, 3), (3, 4), (1, 3), (1, 2), (0, 1)]
        )

        path, path_str = np.einsum_path(*long_test2, optimize="optimal")
        self.assert_path_equal(
            path, ["einsum_path", (0, 5), (1, 4), (3, 4), (1, 3), (1, 2), (0, 1)]
        )

    def test_edge_paths(self):
        # Difficult edge cases

        # Edge test1
        edge_test1 = self.build_operands("eb,cb,fb->cef")
        path, path_str = np.einsum_path(*edge_test1, optimize="greedy")
        self.assert_path_equal(path, ["einsum_path", (0, 2), (0, 1)])

        path, path_str = np.einsum_path(*edge_test1, optimize="optimal")
        self.assert_path_equal(path, ["einsum_path", (0, 2), (0, 1)])

        # Edge test2
        edge_test2 = self.build_operands("dd,fb,be,cdb->cef")
        path, path_str = np.einsum_path(*edge_test2, optimize="greedy")
        self.assert_path_equal(path, ["einsum_path", (0, 3), (0, 1), (0, 1)])

        path, path_str = np.einsum_path(*edge_test2, optimize="optimal")
        self.assert_path_equal(path, ["einsum_path", (0, 3), (0, 1), (0, 1)])

        # Edge test3
        edge_test3 = self.build_operands("bca,cdb,dbf,afc->")
        path, path_str = np.einsum_path(*edge_test3, optimize="greedy")
        self.assert_path_equal(path, ["einsum_path", (1, 2), (0, 2), (0, 1)])

        path, path_str = np.einsum_path(*edge_test3, optimize="optimal")
        self.assert_path_equal(path, ["einsum_path", (1, 2), (0, 2), (0, 1)])

        # Edge test4
        edge_test4 = self.build_operands("dcc,fce,ea,dbf->ab")
        path, path_str = np.einsum_path(*edge_test4, optimize="greedy")
        self.assert_path_equal(path, ["einsum_path", (1, 2), (0, 1), (0, 1)])

        path, path_str = np.einsum_path(*edge_test4, optimize="optimal")
        self.assert_path_equal(path, ["einsum_path", (1, 2), (0, 2), (0, 1)])

        # Edge test5
        edge_test4 = self.build_operands(
            "a,ac,ab,ad,cd,bd,bc->", size_dict={"a": 20, "b": 20, "c": 20, "d": 20}
        )
        path, path_str = np.einsum_path(*edge_test4, optimize="greedy")
        self.assert_path_equal(path, ["einsum_path", (0, 1), (0, 1, 2, 3, 4, 5)])

        path, path_str = np.einsum_path(*edge_test4, optimize="optimal")
        self.assert_path_equal(path, ["einsum_path", (0, 1), (0, 1, 2, 3, 4, 5)])

    def test_path_type_input(self):
        # Test explicit path handling
        path_test = self.build_operands("dcc,fce,ea,dbf->ab")

        path, path_str = np.einsum_path(*path_test, optimize=False)
        self.assert_path_equal(path, ["einsum_path", (0, 1, 2, 3)])

        path, path_str = np.einsum_path(*path_test, optimize=True)
        self.assert_path_equal(path, ["einsum_path", (1, 2), (0, 1), (0, 1)])

        exp_path = ["einsum_path", (0, 2), (0, 2), (0, 1)]
        path, path_str = np.einsum_path(*path_test, optimize=exp_path)
        self.assert_path_equal(path, exp_path)

        # Double check einsum works on the input path
        noopt = np.einsum(*path_test, optimize=False)
        opt = np.einsum(*path_test, optimize=exp_path)
        assert_almost_equal(noopt, opt)

    def test_path_type_input_internal_trace(self):
        # gh-20962
        path_test = self.build_operands("cab,cdd->ab")
        exp_path = ["einsum_path", (1,), (0, 1)]

        path, path_str = np.einsum_path(*path_test, optimize=exp_path)
        self.assert_path_equal(path, exp_path)

        # Double check einsum works on the input path
        noopt = np.einsum(*path_test, optimize=False)
        opt = np.einsum(*path_test, optimize=exp_path)
        assert_almost_equal(noopt, opt)

    def test_path_type_input_invalid(self):
        path_test = self.build_operands("ab,bc,cd,de->ae")
        exp_path = ["einsum_path", (2, 3), (0, 1)]
        assert_raises(RuntimeError, np.einsum, *path_test, optimize=exp_path)
        assert_raises(RuntimeError, np.einsum_path, *path_test, optimize=exp_path)

        path_test = self.build_operands("a,a,a->a")
        exp_path = ["einsum_path", (1,), (0, 1)]
        assert_raises(RuntimeError, np.einsum, *path_test, optimize=exp_path)
        assert_raises(RuntimeError, np.einsum_path, *path_test, optimize=exp_path)

    def test_spaces(self):
        # gh-10794
        arr = np.array([[1]])
        for sp in itertools.product(["", " "], repeat=4):
            # no error for any spacing
            np.einsum("{}...a{}->{}...a{}".format(*sp), arr)


class TestMisc(TestCase):
    def test_overlap(self):
        a = np.arange(9, dtype=int).reshape(3, 3)
        b = np.arange(9, dtype=int).reshape(3, 3)
        d = np.dot(a, b)
        # sanity check
        c = np.einsum("ij,jk->ik", a, b)
        assert_equal(c, d)
        # gh-10080, out overlaps one of the operands
        c = np.einsum("ij,jk->ik", a, b, out=b)
        assert_equal(c, d)


if __name__ == "__main__":
    run_tests()
